use std::{collections::HashMap, path::PathBuf, process::Stdio};

use lazy_static::lazy_static;
use log::Level;
use tokio::sync::{Mutex, RwLock};

use crate::{config::ConfigWrapper, manifest::{song::{Song, SongType}, Format, Manifest}};

#[allow(dead_code)]
#[derive(Debug, Clone)]
struct Proc {
    url: String,
    path: String,
    finished: bool
}

lazy_static!(
    static ref PROCESSES: Mutex<RwLock<HashMap<usize, Proc>>> = Mutex::new(RwLock::new(HashMap::new()));
);

#[derive(Debug, Default)]
pub struct Downloader {
    count: usize,
    nb_initial_song_count: usize,
    nb_cache: Vec<(String, String, Song, Format)>
}

impl Downloader {
    pub fn new() -> Self {
        Self {
            ..Default::default()
        }
    }

    pub fn get_initial_song_count_nb(&self) -> usize {
        self.nb_initial_song_count
    }

    pub fn get_songs_left_nb(&self) -> usize {
        self.nb_cache.len()
    }

    pub fn download_all_nb(&mut self, manifest: &Manifest, cfg: &ConfigWrapper) -> anyhow::Result<usize> {
        for (pname, playlist) in manifest.get_playlists() {
            for (sname, song) in playlist.get_songs() {
                self.nb_cache.push((pname.clone(), sname.clone(), song.clone(), manifest.get_format().clone()));
            }
        }
        self.nb_initial_song_count = self.nb_cache.len();

        self.download_all_nb_poll(cfg)
    }

    pub fn download_all_nb_poll(&mut self, cfg: &ConfigWrapper) -> anyhow::Result<usize> {
        if !crate::process_manager::is_proc_queue_full(10)  {
            if let Some((pname, sname, song, format)) = self.nb_cache.pop() {
                self.download_song(cfg, &sname, &song, &pname, &format)?;
            }
        }

        Ok(crate::process_manager::purge_done_procs())
    }



    pub async fn download_all(&mut self, manifest: &Manifest, cfg: &ConfigWrapper) -> anyhow::Result<usize> {
        let format = manifest.get_format();

        for (name, playlist) in manifest.get_playlists() {
            for (song_name, song) in playlist.get_songs() {
                self.download_song(cfg, song_name, song, &name, format)?;
                self.count += crate::process_manager::wait_for_procs_untill(10)?;
            }
        }
        self.count += crate::process_manager::wait_for_procs_untill(0)?;
        Ok(self.count)
    }
    
    pub fn download_song(&mut self, cfg: &ConfigWrapper, name: &String, song: &Song, playlist: &String, format: &Format) -> anyhow::Result<()> {
        let dl_dir = format!("{}/{playlist}", cfg.cli.output);
        let dl_file = format!("{dl_dir}/{}.{}", name, &format);

        if PathBuf::from(&dl_file).exists() {
            log::debug!("File {dl_file} exists, skipping");
            return Ok(())
        }

        log::debug!("File {dl_file} doesnt exist, downloading");
        let mut cmd = match song.get_type() {

            &SongType::Youtube => {
                log::debug!("Song {} is from yotube", song.get_url_str());
                let mut cmd = tokio::process::Command::new(&cfg.cfg.ytdlp.path);
                cmd.args([
                        "-x",
                        "--audio-format",
                        &format.to_string(),
                        "-o",
                        dl_file.as_str(),
                        song.get_url_str().as_str()
                    ]);
                cmd
            }
            SongType::Spotify => {

                let mut cmd = tokio::process::Command::new(&cfg.cfg.spotdl.path);
                cmd.args([
                    "--format",
                    &format.to_string(),
                    "--output",
                    dl_dir.as_str(),
                    song.get_url_str().as_str()
                ]);
                cmd
            }
            url => {
                log::error!("Unknown or unsupported hostname '{:?}'", url);
                return Ok(());
            }
        };
                
        if log::max_level() < Level::Debug {
            cmd.stdout(Stdio::null()).stderr(Stdio::null());
        };

        crate::process_manager::add_proc(cmd, format!("Downloaded {dl_file}"))?;
        Ok(())
    }
}
